/****************************************************************************
 * Webify - version 2.1				   12/25/95 Steve Ward	    *
 *  with contributions from Robin Duquette				    *
 *     converts Postscript to web-browsable html/jpeg			    *
 *									    *
 * USAGE:								    *
 *   webify <options> <template-file> <postscript-file-name> [subdir]	    *
 *									    *
 *      <template-file> is a template for the generated HTML page,	    *
 *      <postscript-file-name> is the postscript file from which to	    *
 * 		generate JPEGs.						    *
 * 	[subdir] is an optional subdirectory to contain the generated files.*
 * 		It defaults to basname(<postscript-file-name].		    *
 *   <options> may include						    *
 *      -Q #    set quality image (0 is worst; 100 is best; default 75)     *
 * 	-T #	set Thumbnail resolution (default 10)			    *
 * 	-S #	set Slide resolution (default 50)			    *
 * 	-r #	Rotation, 90-degree increments CW (-r 3 usual landscape)    *
 *      -n #    Don't run gs, make pictures; just build n .html pages.      *
 *	-t text	specify title                                               *
 ****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>

#define KEYL 40			/* Max length of a keyword		    */

char *TemplateName,		/* Name of html template file		    */
     *PostscriptName,		/* Name of .ps file, sans .ps suffix	    */
     FullPostscriptName[200],	/* Name of .ps file, with .ps suffix	    */
     *Subdir,			/* Directory name prefix, eg "Lect/"	    */
     *HTMLName,			/* Name of HTML file			    */
     *SubDir,			/* Subdirectory for output files.	    */
     *GSName = "gs",		/* Name of Ghostscript command		    */
     *PostCommandName="cjpeg",	/* Name of post processor command	    */
     *HackPPMName = "hackppm",	/* Name of hackppm command		    */
     Date[16],			/* Today's date				    */
     *PixSuffix,		/* eg, ".jpeg" or ".gif"		    */
     *ProgName;			/* Name of this program ("webify", usually) */

/* Command line options:						    */
char Thumbnails = 1;		/* Use thumbnails.			    */
int  Rotation;			/* Rotation, CW 90-deg increments	    */
char Negative;			/* Negative (reverse video) mode	    */
const int DefaultQualImage = 75;
int  QualityImage;              /* Quality images                           */
int  ThumbRes, DefaultThumbRes = 10;	/* Thumbnail resolution		    */
int  SlideRes, DefaultSlideRes = 50;	/* Slide resolution		    */
char Debug;
int  CommentChar = '|';		/* Default comment character.		    */
int  NoGS;			/* Flag: don't do ghostscript		    */
int  NoCheck;			/* Flag: don't check environment	    */
int  NoFifo;			/* Flag: don't use Unix FIFOs		    */

double CropTop,  CropBot,	/* Optional cropping of page margins	    */
       CropLeft, CropRight;

FILE *InFile, *OutFile;		/* Input template, output HTML files	    */

int  LineNo;			/* Source line #			    */
char *LastKey;			/* Opening escape, for error report.	    */
int   LastKeyLine;		/* Line # of above.			    */

char Title[100];
char BaseName[100];		/* Name of input file, sans suffix.	    */
char PostCommand[200];		/* Post-processor command.		    */

/* Following are used by {for-each-page} and BuildJPegs....		    */
int Pages,			/* Number of pages (JPEG files)		    */
    CurPage;			/* Page we're currently on 		    */

struct OutStream
 { FILE *OutFile;		/* if FILE output			    */
   char *Ptr, *EPtr;		/* if character string output		    */
   struct OutStream *old;	/* Previous output stream.		    */
 } *CurrentOutput;		/* The current output stream.		    */

/* Forward refs								    */
char *FileSize(char *name);
#define ComputeImageName(n,prefix) ComputeSuffixName(n, prefix, PixSuffix)
#define ComputePpmName(n,prefix) ComputeSuffixName(n,prefix,".ppm")
#define ComputeBaseName(n, prefix) ComputeSuffixName(n,prefix,"")
char *ComputeSuffixName(int n, char *prefix, char *suffix);
char *ComputeRelName(int n, char *prefix, char *suffix);
char *NameInSubDir(char *name);
char *ExpandString(char *template, char *target);

void SetJPEG()
 { if (!PostCommandName) PostCommandName = "cjpeg";
   PixSuffix = ".jpeg";
   sprintf(PostCommand, "%s -quality %d -optimize ",
	   PostCommandName, QualityImage);
   strcat(PostCommand, " > %B.jpeg");
 }

void SetGIF()
 { if (!PostCommandName) PostCommandName = "ppmtogif";
   if (!strcmp(PostCommandName, "cjpeg")) PostCommandName = "ppmtogif";
   PixSuffix = ".gif";
   sprintf(PostCommand, "%s ", PostCommandName);
   strcat(PostCommand, " > %B.gif");
 }

char *MAlloc(int size)		/* Our interface to malloc		   */
 { char *it = (char *) malloc(size);
   if (!it) { fprintf(stderr, "webify: out of storage in MAlloc(%d)\n", size);
 	      exit(-1); }
   return it;
 }

char *SCopy(char *s)		/* Copy a string to malloc'ed storage     */
 { int len = strlen(s)+1;
   char *it;
   it = MAlloc(len);
   strcpy(it,s);
   return it;
 }


/* Parse the name of a postscript file.
 * Fills PostscriptName, also used as the default for subdirectory name.
 */

ParsePSName(char *name)
 { char *lastdot=0, *lastslash = 0, *aa, *bb, *base=name;
   static char basename[100];
   for (aa=name; *aa; aa++)
    { if (*aa == '.') lastdot = aa;
      if (*aa == '/') lastslash = aa;
    }
   if (lastslash) base = lastslash+1;
   if (lastdot && strncmp(lastdot, ".ps", 3)) lastdot = 0;
   for (aa=basename, bb=base; *bb && (bb != lastdot);) *aa++ = *bb++;
   PostscriptName = basename;
   SubDir = basename;
   for (aa=FullPostscriptName, bb=name; *bb && (bb != lastdot);)
   	*aa++ = *bb++;
   for (bb=".ps"; *aa++ = *bb++; );
   if (!FileSize(FullPostscriptName))
    { fprintf(stderr, "Can't find postscript input %s\n", FullPostscriptName);
      exit(-1);
    }
}

/* Write output character to current output stream.			    */
WrCh(char ch)
 { if (!CurrentOutput) return;
   if (CurrentOutput->OutFile) putc(ch, CurrentOutput->OutFile);
   if (CurrentOutput->Ptr)
    { *(CurrentOutput->Ptr++) = ch;
      if ((CurrentOutput->Ptr++) >= (CurrentOutput->EPtr))
      	SynErr("String buffer output overflow");
      *(CurrentOutput->Ptr++) = 0;
    } 
 }

WrStr(char *str)
 { if (!str) return WrStr("<null>");
   while (*str) WrCh(*str++);
 }

/* Read next character from input.  Counts lines, ignores comments.	    */
int RdCh()
 { int ch = getc(InFile);
agn:
   if (ch == CommentChar)
    { while (((ch = getc(InFile)) != '\n') && (ch != EOF));
      goto agn;
    }
   switch(ch)
    { case '\n': ++LineNo; return ch;
      case '\\': switch (ch = getc(InFile))
       { case 'n': return '\n';
	 case 't': return '\t';
	 case 'b': return '\b';
	 default: return ch;
       }
      default:	return ch;
    }
 }

UnRdCh(int ch)
 { ungetc(ch, InFile);
 }

/* Fetch a "word" from within current { ... }; returns NULL if end.
 */

char *RdWord()
 { int ch, *p, i, quotes=0;
   static char buf[100];
   while (((ch = RdCh()) != EOF) && (ch <= ' '));
   if (ch == '"') { ++quotes; ch = RdCh(); }

   for (i=0; ; ch = RdCh())
    { if (ch == EOF) break;
      if (!quotes)
       { if (ch == '}') { UnRdCh(ch); break; }
         if (ch <= ' ') break;
       }
      else
       { if (ch == '"') break;
	 if (ch == '\n') SynErr("Unclosed quoted string");
       }
      buf[i++] = ch;
    }
   if (!i) return NULL;
   buf[i] = 0;
   return buf;
 }

DummyKey(char *key)
 { char buf[100];
   sprintf(buf, "{%s ... ignored}", key);
   WrStr(buf);
   return IgnoreEscape();
 }

/* Interpolate text, skip to end of escape				    */
TextKey(char *str)
 { WrStr(str);
   IgnoreEscape();
 }

/* Find file size, return as a char string				    */
#include <sys/stat.h>

char *FileSize(char *name)
 { struct stat buf;
   long size;
   static char sbuf[20];
   if (stat(name, &buf) < 0) return NULL;
   size = buf.st_size;
   if (size < 10000) sprintf(sbuf, "%d", size);
   else if (size < 1000000) sprintf(sbuf, "%dK", (size+512)>>10);
   else { size += 50*1024;
	  sprintf(sbuf, "%d.%dM", (size>>20),
 		 (size & 0xFFFFF) / (100*1024));
	}
   return sbuf;
 }

/* Output an integer, taking the format from the next argument
 * within the current { ... }.
 * Format defaults to "%d".
 */
FormatIntKey(int n)
 { char *fmt = RdWord(), buf[20];
   if (!fmt) fmt = "%d";
   sprintf(buf, fmt, n);
   WrStr(buf);
   IgnoreEscape();
 }

/* Is ch a break character?						    */
int BreakCh(int ch)
 { switch(ch)
    { case '{': case '}': case '(': case ')':
      case ' ':	case '\n': case '\t': case EOF:
      		return 1;
      default: 	return 0;
    }
 }

/* Grab a "word" from the template input stream:			    */
char *GrabWord()
 { static char buf[100];
   int i=0, ch;
   while (((ch = RdCh()) <= ' ') && (ch != EOF));
   buf[i++] = ch; buf[i] = 0;
   if (BreakCh(ch)) return buf;
   while (((ch = RdCh()) > ' ') && (ch != EOF))
    { if (BreakCh(ch))
       { if (i == 0) buf[i++] = ch;
	 else ungetc(ch, InFile);
	 buf[i] = 0;
	 goto rtn;
       }
      buf[i++] = ch;
      buf[i] = 0;
      if (i >= 99) goto rtn;
    }
rtn:
   return buf;
 }

/* Convert a float from input template, return as value
 */

double FloatKey()
 { double x;
   char *cc = GrabWord();
   if (sscanf(cc, "%lf", &x) != 1) SynErr("Bad real number");
   IgnoreEscape();
   return x;
 }

/* Process an escape.
 */
DoKey(char *key)
 { if (!strcmp(key, "comment")) return IgnoreEscape();
   if (!strcmp(key, "quietly")) return Quietly();

   if (!strcmp(key, "title")) return TextKey(Title);
   if (!strcmp(key, "date")) return TextKey(Date);
   if (!strcmp(key, "for-each-page")) return ForEachPage();
   if (!strcmp(key, "thumbnail"))
   	return TextKey(ComputeRelName(CurPage, "T", PixSuffix));
   if (!strcmp(key, "slide"))
   	return TextKey(ComputeRelName(CurPage, "S", PixSuffix));

   if (!strcmp(key, "this-page"))
   	return TextKey(ComputeRelName(CurPage, "P", ".html"));
   if (!strcmp(key, "first-page"))
   	return TextKey(ComputeRelName(1, "P", ".html"));
   if (!strcmp(key, "last-page"))
   	return TextKey(ComputeRelName(Pages, "P", ".html"));
   if (!strcmp(key, "prev-page"))
   	return TextKey(ComputeRelName(CurPage>1? CurPage-1:Pages, "P", ".html"));
   if (!strcmp(key, "next-page"))
   	return TextKey(ComputeRelName(CurPage>=Pages? 1 :CurPage+1, "P",".html"));

   if (!strcmp(key, "page-number")) return FormatIntKey(CurPage);
   if (!strcmp(key, "pages")) return FormatIntKey(Pages);

   if (!strcmp(key, "write-page-file")) return WritePageFile();
   if (!strcmp(key, "write-file")) return WriteFile();

   if (!strcmp(key, "postscript-size"))
   	return TextKey(FileSize(FullPostscriptName));
   if (!strcmp(key, "postscript")) return TextKey(PostscriptName);

   if (!strcmp(key, "rotate"))
    { Rotation = FloatKey(); return; }

   if (!strcmp(key, "crop-top"))
    { CropTop = FloatKey(); return; }
   if (!strcmp(key, "crop-bot"))
    { CropBot = FloatKey(); return; }
   if (!strcmp(key, "crop-left"))
    { CropLeft = FloatKey(); return; }
   if (!strcmp(key, "crop-right"))
    { CropRight = FloatKey(); return; }

   if (!strcmp(key, "slide-resolution"))
    { int temp = (int) FloatKey();
      if (!SlideRes) SlideRes = temp; return; }
   if (!strcmp(key, "thumbnail-resolution"))
    { int temp = (int) FloatKey();
      if (!ThumbRes) ThumbRes = temp; return; }

   SynErr("Unknown escape keyword");
 }

ForEachPage()
 { long seekpt = ftell(InFile);

   if (!Pages) return IgnoreEscape();
   for (CurPage=1; CurPage<=Pages; CurPage++)
    { if (ParseTemplate() != '}') SynErr("Unterminated {for-each-page...}");
      fseek(InFile, seekpt, 0);
    }
   return IgnoreEscape();
 }


/* Ignore input thru closing curly bracket				    */
IgnoreEscape()
 { int ch, count=1;
   for (;;) switch (ch = RdCh())
    { case '{': ++count; continue;
      case '}': if (!--count) return; continue;
      case EOF: SynErr("Missing '}'");
      default:	continue;
    }
 }

/* Flush output, interpret escapes:
 * {quietly ...}
 */
Quietly()
 { struct OutStream out;
   char ch;

   out.old = CurrentOutput;
   out.OutFile = 0;
   out.Ptr = 0;
   CurrentOutput = &out;

   ch = ParseTemplate();
   if (ch != '}') SynErr("Missing '}'");
   CurrentOutput = out.old;
 }

/* Write out a page file:
 * form {write-page-file  ... } writes ... into file named "name".
 */
WritePageFile()
{
    writefile(ComputeRelName(CurPage, "P", ".html"));
    return;
}


/* Write out a file:
 * form {write-file name... } writes ... into file named "name".
 */
WriteFile()
{
    int c;
    char buf[100];
    int i;

    for (;;) {
	switch (c = RdCh()) {
	case ' ': case '\t': case '\n':
	    continue;
	case '}':  SynErr("Premature '}':  missing filename");
	case EOF:  SynErr("Premature EOF:  missing filename");
	}
	break;
    }
    /* Start of filename. */
    i = 0;
    do {
	buf[i++] = c;
	switch (c = RdCh()) {
	case ' ': case '\t': case '\n': case '}':
	    goto out;
	case EOF:  SynErr("Premature EOF in write-file: missing filename");
	}
    } while (i < sizeof(buf));
    SynErr("filename too long");

out:;
    buf[i] = '\0';
    writefile(buf);
}

/* Start writing to file "fn", saving current output. */
writefile(fn)
char *fn;
 { struct OutStream out;
   char ch;
   out.old = CurrentOutput;
   if (!(out.OutFile = fopen(NameInSubDir(fn), "w")))
    { fprintf(stderr, "Can't write file %s\n", fn); exit(-1); }
   out.Ptr = 0;
   CurrentOutput = &out;
   ch = ParseTemplate();
   if (ch != '}') SynErr("Missing '}'");
   fclose(out.OutFile);
   CurrentOutput = out.old;
 }


/* Evaluate text into a string, until next '}'
 * Ptr is start, size is max length.
 */
ParseTo(char *Ptr, int size)
 { struct OutStream out;
   int ch;
   out.OutFile = 0;
   out.old = CurrentOutput;
   CurrentOutput = &out;

   out.Ptr = Ptr;  out.EPtr = Ptr+size;
   ch = ParseTemplate();
   if (ch != '}') SynErr("Missing '}'");
   CurrentOutput = out.old;
   return ch;
 }

/* Read the template file.
 * If pass is zero, just scan and set parameters.
 * If pass is one, write the HTML output file.
 */
DoFile(pass)
 { char *fn;
   int ch;
   struct OutStream out;

   out.OutFile = 0; out.Ptr = 0;
   out.old = CurrentOutput;
   CurrentOutput = &out;

   if (!(InFile = fopen(TemplateName, "r")))
    { fprintf(stderr, "%s: can't read file %s\n", TemplateName); exit(-1); }
   if (pass)
    { if (!(out.OutFile = fopen(fn=NameInSubDir(HTMLName), "w")))
       { fprintf(stderr,
		 "%s: can't write file %s\n", ProgName, fn);
	 exit(-1);
       }
    }

   LineNo = 1;
   if (ParseTemplate() != EOF) SynErr("Premature EOF -- missing }?");
   CurrentOutput = out.old;
   if (out.OutFile) fclose(out.OutFile);
 }

/* Parse the template file.
 * Reads & processes until EOF or '}' read.
 */
int ParseTemplate()
 { int ch;
   while ((ch = RdCh()) != EOF) switch (ch)
    { case '}':	return ch;
      case '{': ParseKey(); continue;
      default:	WrCh(ch);
    }
   return EOF;
 }

/* Parse an escape sequence {key...}.
 * Called after initial { read; returns after closing } read.
 */
ParseKey()
 { char key[KEYL+1], *OldLastKey=LastKey;
   int OldLastKeyLine=LastKeyLine, ch, i;
   for (i=0;;)
    { ch = RdCh();
      if (ch <= ' ') break;
      if (ch == '}') { ungetc(ch, InFile); break; }
      if (i >= KEYL) SynErr("Keyword too long");
      key[i++] = ch;
    }
   LastKey = key;
   LastKeyLine = LineNo;
   key[i] = 0;
   DoKey(key);
   LastKeyLine = OldLastKeyLine;
   LastKey = OldLastKey;
 }


#define NXARG (*++arg? arg : argv[++argn])

main(int argc, char **argv)
{ int argn,i,j;
  char *arg;

  ProgName = argv[0];
  HTMLName = "index.html";
  QualityImage = DefaultQualImage;
  SetJPEG();

  /* Get today's date for {date} escape; thanks to daasche@ee.pdx.edu		*/
  { time_t tloc;
    struct tm *localtime();
    struct tm *tm;
    int status;
    time(&tloc);
    tm = localtime(&tloc);
    strftime(Date, 16, "%D", tm);
  }

  for (argn=1; argn<argc; argn++)
   { if (*(arg=argv[argn]) == '-') 
      { if (!strcmp(arg, "-jpeg")) { SetJPEG(); continue; }
	if (!strcmp(arg, "-gif"))  { SetGIF(); continue; }
	switch (*++arg)
	 { case 'd':	Debug=1; continue;
	   case 'n': 	NoGS = atoi(NXARG);
	   		if (!NoGS) { Usage(); exit(-1); }
	   		continue;
	   case 'r':	Rotation = atoi(NXARG); continue;
	   case 't':	strcpy(Title, NXARG); continue;
	   case 'Q': 	QualityImage = atoi(NXARG);
			if (QualityImage < 0 || QualityImage > 100)
			 { fprintf(stderr,"Webify (warning): the quality parameter ");
			   fprintf(stderr, "must have a value between 0 and 100.\n");
			   fprintf(stderr,
				   "  The quality parameter is set to its default");
			   fprintf(stderr, " value, i.e., %d\n",DefaultQualImage);
			   QualityImage = DefaultQualImage;
			 }
	   		SetJPEG();
			continue;

	   case 'T':	ThumbRes = atoi(NXARG); continue;
	   case 'S':	SlideRes = atoi(NXARG); continue;

	   case 'G':	GSName = NXARG; continue;
	   case 'p':	PostCommandName = NXARG; continue;
	   case 'H':	HackPPMName = NXARG; continue;

	   case 'C':	return CheckEnv();
	   case 'N':	NoCheck = 1; continue;
	   case 'F':	NoFifo = 1; continue;
	   case 'v':	Negative = 1; continue;

	   default:	fprintf(stderr,
				"webify: unknown argument %s\n", argv[argn]);
			Usage();
	 }
      }
    else { 
      if (!TemplateName) TemplateName = arg;
    else 
      if (!PostscriptName) ParsePSName(arg);
    else 
      SubDir = arg;
    }
  }

  if (!TemplateName || !ParsePSName) { Usage(); return 0; }
  if (!NoCheck) CheckEnv();

  DoFile(0);			/* Scanning pass, for parameters */
  
  mkdir(SubDir, 0775);
  link(FullPostscriptName, NameInSubDir(FullPostscriptName));
  
  if (!NoGS) for (i=10,j=1; ; j++)
   { /* Delete existing thumbnails.  */
      if (unlink(NameInSubDir(ComputeRelName(j,"T",PixSuffix))))
      { 
	if (--i <= 0) break; 
      }
      unlink(NameInSubDir(ComputeRelName(j,"S",PixSuffix)));
   }

  DoFile(0);			/* 1st pass thru template file...  */
  if (!NoGS) BuildJPegs();	/* Do the .jpeg conversions */
  else Pages = NoGS;		/* or fudge # of pages...	   */
  DoFile( 1);			/* Do it for real this time.  */
  return 0;
}

SynErr(char *msg)
 { fprintf(stderr, "%s:%d: %s\n", TemplateName, LineNo, msg);
   if (LastKey)
    { fprintf(stderr, "   {%s ... begins on line %d\n",
	      LastKey, LastKeyLine);
    }
   fprintf(stderr, "%s aborted.\n", ProgName);
   exit(-1);
 }

Usage()
 {
   printf("%s: convert Postscript to web-browsable html/jpeg/gif subtree\n",
	  ProgName);
   printf("Version %s.  For latest version & info, see\n", VERSION);
   printf("   http://cag-www.lcs.mit.edu/webify\n");
   printf("USAGE:\n");
   printf("  webify <options> template psfile\n");
   printf("    template is a text file containing prototype HTML and\n");
   printf("      escapes which specify the configuration of generated\n");
   printf("      image files and HTML documents (brief summary below).\n");
   printf("    psfile.ps is the postscript file whose pages are to\n");
   printf("      become web-browsable.\n");
   printf(" <options> may include:\n");
   printf("    -r #  rotate #*90 deg CW; try 3 for landscape.\n");
   printf("    -t \"title\" A title for the presentation.\n");
   printf("    -Q #  Quality of JPEG image (5-95 useful).  Default is 75.\n");
   printf("    -T #  Thumbnail resolution (hence size). Default is 10.\n");
   printf("    -S #  Slide/Page resolution (hence size).  Default is 50.\n");

   printf("    -G <name> Use <name> as command to invoke Ghostscript.  Default '%s'.\n",
	  GSName);
   printf("    -p <name> Use <name> as postprocessing filter (eg cjpeg, ppmtogif).\n");
   printf("    -H <name> Use <name> as command to invoke HACKPPM. Default '%s'.\n",
	  HackPPMName);
   printf("    -F    Don't use UNIX FIFOs.  OS compatibility, at cost of scatch disk space.\n");
   printf("    -n #  Don't run gs or build images; just write # .html pages\n");
   printf("    -d    Debug mode.\n");
   printf("    -C    Check environment for auxilliary programs.\n");
   printf("    -N    Suppress environment check.\n");
   printf("    -v    Reverse Video (Negative) mode.\n");
   printf("    -jpeg Images in jpeg format, using cjpeg (default)\n");
   printf("    -gif  Images in gif format, using ppmtogif\n");

   printf("\nEscapes in template include:\n");
   printf("    {title} - insert title.\n");
   printf("    {date} - insert today's date.\n");
   printf("    {quietly ...} - evaluate for escapes, no output.\n");
   printf("    {crop-top #} Crop # (a float) from top edge.  Similarly for\n");
   printf("        crop-bot, crop-left, crop-right.\n");
   printf("    {slide-resolution #} set slide resolution/size (default 50)\n");
   printf("    {thumbnail-resolution #} set thumbnail resolution/size (default 10)\n");

   printf("    {rotate #} - like -r # command option");

   printf("\n    {for-each-page ...} - repeat ... for each page. May include:\n");
   printf("      {thumbnail} - reference to thumbnail JPEG for current page.\n");
   printf("      {slide} - reference to slide JPEG for current page.\n");
   printf("      {this-page}, {next-page}, {prev-page}, {first-page}, {last-page}\n");
   printf("          - references to per-page HTML files.\n");
   printf("      {write-page-file ...} Put ... in current per-page HTML file.\n");
   printf("      {write-file fn ...} Put ... in file fn.\n");
   printf("      {page-number [fmt]} - numeric page number, formatted per\n");
   printf("          printf format [fmt].  Default is %c%s%c\n", '"', "%d", '"');
   printf("      {pages [fmt]} - total number of pages, formatted per [fmt].\n");

   printf("\nNB: Webify invokes Ghostscript (via gs command).\n");
   printf("    Gs must be installed and have ppmraw conversion enabled.\n");
   printf("    The commands cjpeg and hackppm must also be installed.\n");
 }


/****************************************************************************
 * Interface to Ghostscript: Build the JPEG Images			    *
 ****************************************************************************/

int	GSRes = 50;		/* Device resolution, in pixels/inch	    */
double  XSize=11, YSize=8.5;	/* Page dimensions, in inches.		    */
double	XOrg=0, YOrg=0;		/* Origin of page			    */
char	*FileNPrefix="S";	/* Template for file name.	            */

char tmpfilename[200];		/* Name of temp file 			    */
FILE *tmpout;			/* Temporary file for Ghostscript input     */

/* Prefix SubDir/ to the given file name:				    */
char *NameInSubDir(char *name)
 { static char buf[100];
   sprintf(buf, "%s/%s", SubDir, name);
   return buf;
 }

/* Compute name, relative to subdirectory				    */
char *ComputeRelName(int n, char *prefix, char *suffix)
 { static char buf[100];
   sprintf(buf, "%s%03d%s", prefix, n, suffix);
   return buf;
 }

/* Compute name, relative to upper (invoking) directory.  */
char *ComputeSuffixName(int n, char *prefix, char *suffix)
 { return NameInSubDir(ComputeRelName(n, prefix, suffix));
 }

int CountJPegs()
 { int i;
   for (i=1; ; i++)
    { if (!FileSize(ComputeImageName(i, FileNPrefix))) return i-1;
    } 
 }

BuildThumbnails()
 { if (!ThumbRes) ThumbRes = DefaultThumbRes;
   GSRes = ThumbRes;
   RunGS("T", ThumbRes);
 }

BuildSlides()
{  if (!SlideRes) SlideRes = DefaultSlideRes;
   RunGS("S", SlideRes);
 }


BuildJPegs()
 {
   if (Debug)
     fprintf(stderr,
	     "Rotation=%d, Crops = %f, %f, %f, %f\n",
	     Rotation, CropTop, CropBot, CropLeft, CropRight);
   BuildThumbnails();
   Pages = CountJPegs();
   if (!Pages)
    { fprintf(stderr,
	      "%s: Ghostscript (%s) command generated no output.\n",
	      ProgName, GSName);
      fprintf(stderr,
	      "  To debug the problem, try viewing your postscript file directly\n");
      fprintf(stderr,
	      "  using the command %c%s %s%c\n", '"', GSName, FullPostscriptName, '"');
      exit(-1);
    }
   fprintf(stderr, "%s: %d pages of %s's being created.\n", ProgName, Pages, PixSuffix);
   BuildSlides();
 }

RunGS(char *NamePrefix, int res)
{ char command[400], cjpeg[400], hackppm[100];
  int i, err, pid;
  static char *gsoutflag = ">";
  FileNPrefix = NamePrefix;

  sprintf(hackppm,
	  "%s rotate % d crop %d %d %d %d %s",
	  HackPPMName,
	  Rotation,
	  (int) (CropTop*res), (int) (CropBot*res),
	  (int) (CropLeft*res), (int)(CropRight*res),
	  Negative? "negative" : "");

  /*
   * Strategy (thanks, Robin):
   * Use UNIX FIFOs, where we can, to reduce scatch disk consumption
   *    by pipelining PPM file creation & consumption.
   * HOWEVER, we can't us FIFOs for thumbnails, since we don't know
   *    how many there are.  MOREOVER, we avoid FIFOs on systems
   *    which don't seem to support them compatibly.
   */
  if (NoFifo || !strcmp(NamePrefix,"T"))	/* Thumbnails: avoid FIFOs */
   { if (Debug) fprintf(stderr, "RunGS('%s', %d)\n", NamePrefix, res);
     sprintf(command,
	    "%s -dNOPAUSE -r%dx%d %s%s/%s%%03d.ppm %s - %sgs.out 2>&1 </dev/null",
	    GSName, res, res,
	    "-sDEVICE=ppmraw -sOutputFile=",
	    SubDir,
	    NamePrefix,
	    FullPostscriptName, gsoutflag);
     gsoutflag = ">>";		/* Next time, append.			*/
     if (Debug) fprintf(stderr, "system(%s) ...\n", command);
     err = system(command);
     if (Debug) fprintf(stderr, "  ... system(...) returns %d\n", err);

    /* Convert PPM files into image files. */
    for (i=1; ; i++)
     { char *PPMFileName = SCopy(ComputePpmName(i, NamePrefix));;
       if (!FileSize(PPMFileName)) break;
       strcpy(BaseName, ComputeBaseName(i, NamePrefix));
	{ char buf[200];
	  sprintf(cjpeg, "%s <%s | ", hackppm, PPMFileName);
	  ExpandString(PostCommand,buf);
	  strcat(cjpeg, buf);
	}
       err = system(cjpeg);
       if (Debug) fprintf(stderr, "system(%s) returns %d\n", cjpeg, err);
       free(PPMFileName);
       if (!Debug) unlink(PPMFileName);
     }
  }
  else					 /* We build the slides. */
   { /* First, we make the fifos. */
     for (i = 1; i <= Pages; i++)
      { char *PPMFileName;
	const mode_t PERMS = 0666;
	PPMFileName = ComputePpmName(i, NamePrefix);
	unlink(PPMFileName);
	if ((mknod(PPMFileName, S_IFIFO | PERMS, 0)) < 0 )
	 { perror("mknod");
	   exit(-1);
	 }
      }
    switch (pid=fork())
     { case -1:	perror("fork");
       		exit(-1);
       		break;

       case 0:
       		/* Convert PPM files into image files. */
       		for (i=1; i <= Pages; i++)
		 { char *PPMFileName;
		   PPMFileName = SCopy(ComputePpmName(i, NamePrefix));
		   strcpy(BaseName, ComputeBaseName(i, NamePrefix));
		   { char buf[200];
		     sprintf(cjpeg, "%s <%s | ", hackppm, PPMFileName);
		     ExpandString(PostCommand,buf);
		     strcat(cjpeg, buf);
		   }
		   err = system(cjpeg);
		   if (Debug) fprintf(stderr,
				      "system(%s) returns %d\n", cjpeg, err);
		   if (!Debug) unlink(BaseName);
		   free(PPMFileName);
		 }
       		exit(0);


       default:
       	if (Debug) fprintf(stderr, "RunGS('%s', %d)\n", NamePrefix, res);
       	sprintf(command,
		"%s -dNOPAUSE -r%dx%d %s%s/%s%%03d.ppm %s - %sgs.out 2>&1 </dev/null",
		GSName, res, res,
		"-sDEVICE=ppmraw -sOutputFile=",
		SubDir,
		NamePrefix,
		FullPostscriptName, gsoutflag);
       if (Debug) fprintf(stderr, "system(%s) ...\n", command);
       err = system(command);
       if (Debug) fprintf(stderr, "  ... system(...) returns %d\n", err);
       while (wait(NULL) != pid);
     }
   }
 }

/****************************************************************************
 * Check local environment, see if we've got a prayer of working...	    *
 ****************************************************************************/

#define TEMPNAME "webify.temp"		/* Temp file for output.	    */

/* Run program, give it args, look for string in its output:		    */

CheckPgm(char *program, char *args)
 { char buf[200]; int status;
   sprintf(buf, "%s %s >%s 2>&1", program, args, TEMPNAME);

   if (Debug) fprintf(stderr, "System(%s)..." , buf);
   status = system(buf);
   if (Debug) fprintf(stderr, " returns 0x%x\n", status);

   return 0xFF & (status>>8); 
 }

GrepResult(char *result)
 { char buf[200]; int status;
   sprintf(buf, "grep %c%s%c %s >/dev/null", '"', result, '"', TEMPNAME);
   if (Debug) fprintf(stderr, "System(%s)..." , buf);
   status = system(buf);
   if (Debug) fprintf(stderr, " returns 0x%x\n", status);
   return status & 0xFF00;
 }

CheckEnv()
 { int errs = 0;

   if (CheckPgm(GSName, "-h"))
    { fprintf(stderr, "Can't run Ghostscript (%s command)...", GSName);
      fprintf(stderr, " is it properly installed?\n");
      unlink(TEMPNAME);
      exit(-1);
    }

   if (1)
    { CheckPgm(PostCommandName, "-h");	
      /* Unfortunately, postprocessor doesn't return appropriate status here... */
      if (GrepResult(PostCommandName))
       { fprintf(stderr, "webify: Unrecognized output from %s -h.\n",
		 PostCommandName);
	 fprintf(stderr, "  Are you sure its installed properly?\n");
	 ++errs;
       }
    }

   if (CheckPgm(HackPPMName, "-h"))
    { fprintf(stderr, "Can't execute hackppm command '%s -h'.\n", HackPPMName);
      fprintf(stderr, "  Is %s installed properly?\n", HackPPMName);
      unlink(TEMPNAME);
      exit(-1);
    }

   if (GrepResult(VERSION))
    { fprintf(stderr, "Seems like wrong hackppm version is installed.\n");
      fprintf(stderr, "Webify version %s requires hackppm version %s\n",
	      VERSION, VERSION);
      unlink(TEMPNAME);
      exit(-1);
    }

   if (errs)
    { fprintf(stderr, "webify WARNING: your envirnonment may not be");
      fprintf(stderr, " appropriately for webify!\n");
    }

   unlink(TEMPNAME);
   return errs;
 }

/* Interpolate a string, expanding %x excapes:
 *   %Q		Quality (-Q # option)
 *   %B		Base filename
 */

char *ExpandString(char *template, char *target)
 { char buf[100], *p=target, *aa;
   while (*template)
    { switch (*template)
       { case '%': switch (*++template)
	  { case 'Q': { char buf[20];
			sprintf(buf, "%d", QualityImage);
			for (aa=buf; *target=*aa++; ++target);
		      }
	 	      ++template; continue;
	    case 'B': for (aa=BaseName; *target=*aa++; ++target);
	 	      ++template; continue;
	    default:  *target++ = *template++; continue;
	  }
	 case 0: goto done;
	 default:  *target++ = *template++;
	 	   continue;
       }
    }
done:
   *target = 0;
 }
